import functools
import sys

import configargparse

from smilelogging.logger import Logger
from smilelogging.slutils import blue, green, red, yellow, bold, update_args


argparser = configargparse.ArgumentParser()
argparser.add_argument(
    "--experiment_name", type=str, default="", help="Experiment name."
)
argparser.add_argument(
    "--debug",
    action="store_true",
    help="All the logs will be saved to `Debug_Dir`.",
)
argparser.add_argument(
    "--resume_expid",
    type=str,
    default="",
    help="The expid used to uniquely identify an experiment.",
)


def warn_deprecated_args(old, new):
    if old in sys.argv:
        print(
            f"[Smilelogging Error] {old} is deprecated now, please use {new} instead and rerun"
        )
        exit(0)


def add_update_args(fn):
    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        ret = fn(*args, **kwargs)
        ret = update_args(ret)
        return ret

    return wrapper


warn_deprecated_args("--project_name", "--experiment_name")
warn_deprecated_args("--project", "--experiment_name")
argparser.parse_args = add_update_args(argparser.parse_args)



import copy
import os
import re
import socket

from colorama import (  # Need to "pip install colorama" first. Compatible with Windows users.
    Back,
    Fore,
    Style,
    init,
)

init()  # Initialize colorama.


def get_exp_name_id(exp_path):
    """arg examples: Experiments/kd-vgg13vgg8-cifar100-Temp40_SERVER5-20200727-220318
            or kd-vgg13vgg8-cifar100-Temp40_SERVER5-20200727-220318
            or Experiments/kd-vgg13vgg8-cifar100-Temp40_SERVER5-20200727-220318/weights/ckpt.pth
            or PruneCnst_AlignCnst__PR0.875__ddprun_RANK0-SERVER196-20230614-194146
    Args:
        exp_path: experiment path (can be the experiment folder or a ckpt)
    Returns:
        ExpID: SERVER5-20200727-220318
        expid: 220318
        expname: kd-vgg13vgg8-cifar100-Temp40
        date: 20200727
    """
    seps = exp_path.split(os.sep)
    try:
        for s in seps:
            if "SERVER" in s:
                expid = s.split("-")[-1]
                assert expid.isdigit()
                ExpID = "SERVER" + s.split("SERVER")[1]
                expname = (
                    s.split("-SERVER")[0] if "_RANK" in s else s.split("_SERVER")[0]
                )
                date = s.split("-")[-2]
    except:
        print(f'Failed to parse "{exp_path}", please check')
        exit(0)
    return ExpID, expid, expname, date


def clean_colored_text(logs):
    # Define the pattern to match the escape sequences
    pattern = r"\x1b\[[0-9;]*[m]"

    # Use regex to find and remove escape sequences from the logs
    cleaned_logs = re.sub(pattern, "", logs)

    return cleaned_logs


def standardize_metricline(line):
    """Make metric line in standard form."""
    line = clean_colored_text(line)
    for m in [
        "(",
        ")",
        "[",
        "]",
        "<",
        ">",
        "|",
        ",",
        ";",
        "!",
        "?",
    ]:  # Some non-numerical, no-meaning marks
        if m in line:
            line = line.replace(m, f" {m} ")
    if ":" in line:
        line = line.replace(":", " ")
    line = " ".join(line.split())
    return line


def get_value(line, key, type_func=float):
    """Get the value of a <key> in <line> in a log txt."""
    # Preprocessing to deal with some non-standard log format
    line = standardize_metricline(line)

    value = line.split(f" {key} ")[1].strip().split()[0]

    # Manually fix some problems.
    if "/" in value:  # E.g., Epoch 199/200
        value = value.split("/")[0]

    if value.endswith("%"):
        value = type_func(value[:-1]) / 100.0
    else:
        value = type_func(value)
    return value


def replace_value(line, key, new_value):
    line = standardize_metricline(line)
    value = line.split(key)[1].strip().split()[0]
    line = line.replace(f" {key} {value} ", f" {key} {new_value} ")
    return line


def get_project_name():
    cwd = os.getcwd()
    # assert '/Projects/' in cwd
    return cwd.split(os.sep)[-1]


# acc line example: Acc1 71.1200 Acc5 90.3800 Epoch 840 (after update) lr 5.0000000000000016e-05 (Best_Acc1 71.3500 @ Epoch 817)
# acc line example: Acc1 0.9195 @ Step 46600 (Best = 0.9208 @ Step 38200) lr 0.0001
# acc line example: ==> test acc = 0.7156 @ step 80000 (best = 0.7240 @ step 21300)
def is_metric_line(line, mark=""):
    """This function determines if a line is an accuracy line. Of course the accuracy line should meet some
    format features which @mst used. So if these format features are changed, this func may not work.
    """
    line = standardize_metricline(line)
    if mark:
        return mark in line
    else:
        line = line.lower()
        return (
            "acc" in line
            and "best" in line
            and "lr" in line
            and "resume" not in line
            and "finetune" not in line
        )


def parse_metric(line, metric, scale=1.0):
    """Parse out the metric value of interest."""
    line = line.strip()
    # Get the last metric
    try:
        metric_l = get_value(line, metric)
    except:
        print(f'Parsing last metric failed; please check! The line is "{line}"')
        exit(1)

    # Get the best metric
    try:
        if f"Best {metric}" in line:  # previous impl.
            metric_b = get_value(line, f"Best {metric}")
        elif f"Best_{metric}" in line:
            metric_b = get_value(line, f"Best_{metric}")
        elif f"Best{metric}" in line:
            metric_b = get_value(line, f"Best{metric}")
        else:
            metric_b = -1  # Not found the best metric value (not written in log)
    except:
        print(f'Parsing best metric failed; please check! The line is "{line}"')
        exit(1)
    return metric_l * scale, metric_b * scale


def parse_time(line):
    """Parse the time (e.g., epochs or steps) in a metric line."""
    line = standardize_metricline(line)
    if " Epoch " in line:
        time = get_value(line, "Epoch", type_func=int)
    elif " Step " in line:
        time = get_value(line, "Step", type_func=int)
    elif " step " in line:
        time = get_value(line, "step", type_func=int)
    else:
        print(f'Fn "parse_time" failed. Please check')
        raise NotImplementedError
    return time


def parse_finish_time(log_f):
    lines = open(log_f, "r").readlines()
    for k in range(1, min(1000, len(lines))):
        line = lines[-k].lower()
        if "(speed:" in line and "per timing, total_time:" in line:
            finish_time = (
                lines[-k].split("(speed:")[0].split()[-1].strip()
            )  # example: predicted finish time: 2020/10/25-08:21 (speed: 314.98s per timing)
            return finish_time


def get_ip():
    # Get IP address. Refer to: https://stackoverflow.com/questions/166506/finding-local-ip-addresses-using-pythons-stdlib
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(("8.8.8.8", 80))
    ip = s.getsockname()[0]
    s.close()
    return ip


class EmptyClass:
    pass


def update_args(args):
    """Update arguments of configargparse"""
    arg_dict = copy.deepcopy(args.__dict__)
    for k, v in arg_dict.items():
        if "." in k:  # TODO-@mst: hardcode pattern, may be risky
            module, arg = k.split(".")  # e.g., "deepmixup.depth"
            if arg_dict[f"{module}.ON"]:  # this module is being used
                if not hasattr(args, module):
                    args.__setattr__(module, EmptyClass())  # set to a blank class
                args.__dict__[module].__dict__[
                    arg
                ] = v  # args.'deepmixup.depth' = 10 --> args.deepmixup.depth = 10
            args.__delattr__(k)
    return args


def red(*msg, sep=","):
    """Wrap log string with red colo"""
    msg = sep.join([str(x) for x in msg])
    return Fore.RED + msg + Style.RESET_ALL


def green(*msg, sep=","):
    """Wrap log string with green colo"""
    msg = sep.join([str(x) for x in msg])
    return Fore.GREEN + msg + Style.RESET_ALL


def yellow(*msg, sep=","):
    """Wrap log string with yellow colo"""
    msg = sep.join([str(x) for x in msg])
    return Fore.YELLOW + msg + Style.RESET_ALL


def blue(*msg, sep=","):
    """Wrap log string with blue colo"""
    msg = sep.join([str(x) for x in msg])
    return Fore.BLUE + msg + Style.RESET_ALL


def bold(*msg, sep=","):
    """Wrap log string in bold"""
    msg = sep.join([str(x) for x in msg])
    return Style.BRIGHT + msg + Style.RESET_ALL



import argparse
from fnmatch import fnmatch
import glob
import os
from datetime import datetime
import pytz

from .slutils import get_exp_name_id, yellow

tz = pytz.timezone("US/Eastern")
today = datetime.now(tz).strftime("*-%Y%m%d-*")

parser = argparse.ArgumentParser()
parser.add_argument(
    "--kw", type=str, default=today, help="keyword for filtering experiment folders"
)
parser.add_argument(
    "--exact_kw",
    action="store_true",
    help="if true, not filter by expname but exactly the kw",
)
parser.add_argument(
    "--metricline_mark",
    type=str,
    default="",
    help="mark to identify which log lines have metric results like accuracy",
)
parser.add_argument("--metric", type=str, default="Acc1")
parser.add_argument(
    "--lastline_mark",
    type=str,
    default="last",
    help="mark to identify which log lines are the last; used to print last metric",
)
parser.add_argument("--remove_outlier", action="store_true")
parser.add_argument(
    "--outlier_thresh",
    type=float,
    default=0.5,
    help="if |value - mean| > outlier_thresh, we take this value as an outlier",
)
parser.add_argument("--ignore", type=str, default="", help="seperated by comma")
parser.add_argument("--exps_folder", type=str, default="Experiments")
parser.add_argument("--n_decimals", type=int, default=4)
parser.add_argument("--scale", type=float, default=1.0)
parser.add_argument("--acc_analysis", action="store_true")
parser.add_argument("--out_plot_path", type=str, default="plot.jpg")
args = parser.parse_args()

print(yellow(args.__dict__), "\n")

# 1st filtering: get all the exps with the keyword
all_exps = [
    x
    for x in glob.glob(f"{args.exps_folder}/{args.kw}")
    if os.path.isdir(x) and "SERVER" in x
]
if len(all_exps) == 0:
    print(f"!! [Warning] Found NO experiments with the given keyword. Please check")

# 2nd filtering: remove all exps in args.ignore
if args.ignore:
    ignores = args.ignore.split(",")
    all_exps = [e for e in all_exps if True not in [fnmatch(e, i) for i in ignores]]

# 3rd filtering: add all the exps with the same name, even it is not included by the 1st filtering by kw
if args.exact_kw:
    exp_groups = {}  # Get group exps, because each group is made up of multiple times
    for exp in all_exps:
        _, _, expname, _ = get_exp_name_id(exp)
        if expname not in exp_groups:
            exp_groups.append(expname)
else:
    exp_groups = {}
    for exp in all_exps:
        _, _, expname, _ = get_exp_name_id(exp)
        if expname not in exp_groups:
            search_pattern = (
                f"{args.exps_folder}/{expname}-SERVER*"
                if "RANK" in expname
                else f"{args.exps_folder}/{expname}_SERVER*"
            )
            exps_with_same_expname = glob.glob(search_pattern)
            exps_with_same_expname.sort()
            exp_groups[expname] = exps_with_same_expname



import time
import os
import copy
from collections import OrderedDict
import glob
import pickle
import subprocess
import functools

import numpy as np
from PIL import Image

import torch
import torch.nn as nn
import torch.nn.init as init
from torch.utils.data import Dataset
import torch.nn.functional as F
from torch.autograd import Variable


def _weights_init(m):
    if isinstance(m, (nn.Conv2d, nn.Linear)):
        init.kaiming_normal(m.weight)
        if m.bias is not None:
            m.bias.data.fill_(0)
    elif isinstance(m, nn.BatchNorm2d):
        if m.weight is not None:
            m.weight.data.fill_(1.0)
            m.bias.data.zero_()


def _weights_init_orthogonal(m, act="relu", scale=1):
    if isinstance(m, (nn.Conv2d, nn.Linear)):
        init.orthogonal_(m.weight, gain=init.calculate_gain(act))
        m.weight.data.copy_(m.weight.data * scale)
        if m.bias is not None:
            m.bias.data.fill_(0)
    elif isinstance(m, nn.BatchNorm2d):
        if m.weight is not None:
            m.weight.data.fill_(1.0)
            m.bias.data.zero_()


# Modify the orthogonal initialization
# refer to: https://pytorch.org/docs/stable/_modules/torch/nn/init.html#orthogonal_
def orthogonalize_weights(tensor, act):
    r"""Fills the input `Tensor` with a (semi) orthogonal matrix, as
    described in `Exact solutions to the nonlinear dynamics of learning in deep
    linear neural networks` - Saxe, A. et al. (2013). The input tensor must have
    at least 2 dimensions, and for tensors with more than 2 dimensions the
    trailing dimensions are flattened.

    Args:
        tensor: an n-dimensional `torch.Tensor`, where :math:`n \geq 2`
        gain: optional scaling factor

    Examples:
        >>> w = torch.empty(3, 5)
        >>> nn.init.orthogonal_(w)
    """
    tensor = tensor.clone().detach()  # @mst: avoid modifying the original tensor
    gain = init.calculate_gain(act)

    if tensor.ndimension() < 2:
        raise ValueError("Only tensors with 2 or more dimensions are supported")

    rows = tensor.size(0)
    cols = tensor.numel() // rows
    # flattened = tensor.new(rows, cols).normal_(0, 1)
    flattened = tensor.view(rows, cols)  # @mst: do NOT reinit the tensor

    if rows < cols:
        flattened.t_()

    # Compute the qr factorization
    q, r = torch.qr(flattened)  # q: (rows, cols), r: (cols, cols) if rows > cols.

    # Make Q uniform according to https://arxiv.org/pdf/math-ph/0609050.pdf
    d = torch.diag(r, 0)  # d: (cols)
    ph = d.sign()
    q *= ph  # (rows, cols) * (cols) => (rows, cols)

    if rows < cols:
        q.t_()

    with torch.no_grad():
        tensor.view_as(q).copy_(q)
        tensor.mul_(gain)
    return tensor


# refer to: https://github.com/JiJingYu/delta_orthogonal_init_pytorch/blob/master/demo.py
def genOrthgonal(dim):
    a = torch.zeros((dim, dim)).normal_(0, 1)
    q, r = torch.qr(a)
    d = torch.diag(r, 0).sign()
    diag_size = d.size(0)
    d_exp = d.view(1, diag_size).expand(diag_size, diag_size)
    q.mul_(d_exp)
    return q


def delta_orthogonalize_weights(weights, act):
    weights = copy.deepcopy(weights)
    gain = init.calculate_gain(act)

    rows = weights.size(0)
    cols = weights.size(1)
    if rows > cols:
        print("In_filters should not be greater than out_filters.")
    weights.data.fill_(0)
    dim = max(rows, cols)
    q = genOrthgonal(dim)
    mid1 = weights.size(2) // 2
    mid2 = weights.size(3) // 2
    weights[:, :, mid1, mid2] = q[: weights.size(0), : weights.size(1)]
    weights.mul_(gain)
    return weights


# refer to: https://github.com/Eric-mingjie/rethinking-network-pruning/blob/master/imagenet/l1-norm-pruning/compute_flops.py
def get_n_params(model):
    total = sum(
        [param.nelement() if param.requires_grad else 0 for param in model.parameters()]
    )
    total /= 1e6
    return total


# The above 'get_n_params' requires 'param.requires_grad' to be true. In KD, for the teacher, this is not the case.
def get_n_params_(model, sparse=False):
    n_params = 0
    LEARNABLES = (
        nn.Conv1d,
        nn.Conv2d,
        nn.Conv3d,
        nn.Linear,
    )  # Only consider Conv2d and Linear, no BN
    print(
        f"The following learnable_layers are accounted for: {[x.__name__ for x in LEARNABLES]}"
    )
    for _, module in model.named_modules():
        if isinstance(module, LEARNABLES):
            n_params += module.weight.numel()
            if hasattr(module, "bias") and type(module.bias) != type(None):
                n_params += module.bias.numel()
    return n_params


def get_n_flops(model=None, input_res=224, multiply_adds=True, n_channel=3):
    model = copy.deepcopy(model)

    prods = {}

    def save_hook(name):
        def hook_per(self, input, output):
            prods[name] = np.prod(input[0].shape)

        return hook_per

    list_1 = []

    def simple_hook(self, input, output):
        list_1.append(np.prod(input[0].shape))

    list_2 = {}

    def simple_hook2(self, input, output):
        list_2["names"] = np.prod(input[0].shape)

    list_conv = []

    def conv_hook(self, input, output):
        batch_size, input_channels, input_height, input_width = input[0].size()
        output_channels, output_height, output_width = output[0].size()

        kernel_ops = (
            self.kernel_size[0] * self.kernel_size[1] * (self.in_channels / self.groups)
        )
        bias_ops = 0 if self.bias is not None else 0

        # params = output_channels * (kernel_ops + bias_ops) # @mst: commented since not used
        # flops = (kernel_ops * (2 if multiply_adds else 1) + bias_ops) * output_channels * output_height * output_width * batch_size

        num_weight_params = (
            (self.weight.data != 0).float().sum()
        )  # @mst: this should be considering the pruned model
        # could be problematic if some weights happen to be 0.
        flops = (
            (
                num_weight_params * (2 if multiply_adds else 1)
                + bias_ops * output_channels
            )
            * output_height
            * output_width
            * batch_size
        )

        list_conv.append(flops)

    list_linear = []

    def linear_hook(self, input, output):
        batch_size = input[0].size(0) if input[0].dim() == 2 else 1

        weight_ops = self.weight.nelement() * (2 if multiply_adds else 1)
        bias_ops = self.bias.nelement()

        flops = batch_size * (weight_ops + bias_ops)
        list_linear.append(flops)

    list_bn = []

    def bn_hook(self, input, output):
        list_bn.append(input[0].nelement() * 2)

    list_relu = []

    def relu_hook(self, input, output):
        list_relu.append(input[0].nelement())

    list_pooling = []

    def pooling_hook(self, input, output):
        batch_size, input_channels, input_height, input_width = input[0].size()
        output_channels, output_height, output_width = output[0].size()

        kernel_ops = self.kernel_size * self.kernel_size
        bias_ops = 0
        params = 0
        flops = (
            (kernel_ops + bias_ops)
            * output_channels
            * output_height
            * output_width
            * batch_size
        )

        list_pooling.append(flops)

    list_upsample = []

    # For bilinear upsample
    def upsample_hook(self, input, output):
        batch_size, input_channels, input_height, input_width = input[0].size()
        output_channels, output_height, output_width = output[0].size()

        flops = output_height * output_width * output_channels * batch_size * 12
        list_upsample.append(flops)

    def foo(net):
        childrens = list(net.children())
        if not childrens:
            if isinstance(net, torch.nn.Conv2d):
                net.register_forward_hook(conv_hook)
            if isinstance(net, torch.nn.Linear):
                net.register_forward_hook(linear_hook)
            if isinstance(net, torch.nn.BatchNorm2d):
                net.register_forward_hook(bn_hook)
            if isinstance(net, torch.nn.ReLU):
                net.register_forward_hook(relu_hook)
            if isinstance(net, torch.nn.MaxPool2d) or isinstance(
                net, torch.nn.AvgPool2d
            ):
                net.register_forward_hook(pooling_hook)
            if isinstance(net, torch.nn.Upsample):
                net.register_forward_hook(upsample_hook)
            return
        for c in childrens:
            foo(c)

    if model == None:
        model = torchvision.models.alexnet()
    foo(model)
    input = Variable(
        torch.rand(n_channel, input_res, input_res).unsqueeze(0), requires_grad=True
    )
    out = model(input)

    total_flops = sum(list_conv) + sum(
        list_linear
    )  # + sum(list_bn) + sum(list_relu) + sum(list_pooling) + sum(list_upsample))
    total_flops /= 1e9
    # print('  Number of FLOPs: %.2fG' % total_flops)

    return total_flops


# The above version is redundant. Get a neat version as follow.
def get_n_flops_(
    model=None, img_size=(224, 224), n_channel=3, count_adds=True, input=None, **kwargs
):
    """Only count the FLOPs of conv and linear learnable_layers (no BN learnable_layers etc.).
    Only count the weight computation (bias not included since it is negligible)
    """
    if hasattr(img_size, "__len__"):
        height, width = img_size
    else:
        assert isinstance(img_size, int)
        height, width = img_size, img_size

    # model = copy.deepcopy(model)
    list_conv = []

    def conv_hook(self, input, output):
        flops = (
            np.prod(self.weight.data.shape)
            * output.size(2)
            * output.size(3)
            / self.groups
        )
        list_conv.append(flops)

    list_linear = []

    def linear_hook(self, input, output):
        flops = np.prod(self.weight.data.shape)
        list_linear.append(flops)

    def register_hooks(net, hooks):
        childrens = list(net.children())
        if not childrens:
            if isinstance(net, torch.nn.Conv2d):
                h = net.register_forward_hook(conv_hook)
                hooks += [h]
            if isinstance(net, torch.nn.Linear):
                h = net.register_forward_hook(linear_hook)
                hooks += [h]
            return

        for c in childrens:
            register_hooks(c, hooks)

    hooks = []
    register_hooks(model, hooks)
    if input is None:
        input = torch.rand(1, n_channel, height, width)
        use_cuda = next(model.parameters()).is_cuda
        if use_cuda:
            input = input.cuda()

    # forward
    is_train = model.training
    model.eval()
    with torch.no_grad():
        model(input, **kwargs)
    total_flops = sum(list_conv) + sum(list_linear)
    if count_adds:
        total_flops *= 2

    # reset to original model
    for h in hooks:
        h.remove()  # clear hooks
    if is_train:
        model.train()
    return total_flops


# refer to: https://github.com/alecwangcq/EigenDamage-Pytorch/blob/master/utils/common_utils.py
class PresetLRScheduler(object):
    """Using a manually designed learning rate schedule rules."""

    def __init__(self, decay_schedule):
        if not isinstance(decay_schedule, dict):
            assert isinstance(decay_schedule, str)
            decay_schedule = strdict_to_dict(decay_schedule, float)

        # decay_schedule is a dictionary
        # which is for specifying iteration -> lr
        self.decay_schedule = OrderedDict()
        for (
            k,
            v,
        ) in (
            decay_schedule.items()
        ):  # a dict, example: {"0":0.001, "30":0.00001, "45":0.000001}
            self.decay_schedule[int(float(k))] = v  # to float first in case of '1e3'
        # print('Using a preset learning rate schedule:')
        # print(self.decay_schedule)

    def __call__(self, optimizer, e):
        epochs = list(self.decay_schedule.keys())
        epochs = sorted(epochs)  # example: [0, 30, 45]
        lr = self.decay_schedule[epochs[-1]]
        for i in range(len(epochs) - 1):
            if epochs[i] <= e < epochs[i + 1]:
                lr = self.decay_schedule[epochs[i]]
                break
        for param_group in optimizer.param_groups:
            param_group["lr"] = lr
        return lr


def get_lr(optimizer):
    for param_group in optimizer.param_groups:
        lr = param_group["lr"]
        return lr


def plot_weights_heatmap(weights, out_path):
    """
    weights: [N, C, H, W]. Torch tensor
    averaged in dim H, W so that we get a 2-dim color map of size [N, C]
    """
    w_abs = weights.abs()
    w_abs = w_abs.data.cpu().numpy()

    fig, ax = plt.subplots()
    im = ax.imshow(w_abs, cmap="jet")

    # make a beautiful colorbar
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size=0.05, pad=0.05)
    fig.colorbar(im, cax=cax, orientation="vertical")

    ax.set_xlabel("Channel")
    ax.set_ylabel("Filter")
    fig.savefig(out_path, dpi=200)
    plt.close(fig)


def strlist_to_list(sstr, ttype=float):
    r"""Example:
    # self.args.stage_pr = [0, 0.3, 0.3, 0.3, 0, ]
    # self.args.skip_layers = ['1.0', '2.0', '2.3', '3.0', '3.5', ]
    turn these into a list of <ttype> (float or str or int etc.)
    """
    if not sstr:
        return sstr
    out = []
    sstr = sstr.strip()
    if sstr.startswith("[") and sstr.endswith("]"):
        sstr = sstr[1:-1]
    for x in sstr.split(","):
        x = x.strip()
        if x:
            x = ttype(x)
            out.append(x)
    return out


def strdict_to_dict(sstr, ttype=float):
    r"""Example: '{"1": 0.04, "2": 0.04, "4": 0.03, "5": 0.02, "7": 0.03}'"""
    if not sstr:
        return sstr
    out = OrderedDict()
    sstr = sstr.strip()
    if sstr.startswith("{") and sstr.endswith("}"):
        sstr = sstr[1:-1]

    sep = ","
    if "/" in sstr:
        sep = "/"
    elif ";" in sstr:
        sep = ";"
    for x in sstr.split(sep):
        x = x.strip()
        if x:
            k = x.split(":")[0]  # note: key is always str
            if k.startswith("'"):
                k = k.strip("'")  # remove ' '
            if k.startswith('"'):
                k = k.strip('"')  # remove " "
            v = ttype(x.split(":")[1].strip())
            out[k] = v
    return out


def check_path(x):
    if x:
        complete_path = glob.glob(x)
        if len(complete_path) > 1:
            raise ValueError(
                f"The given path ({x}) points to multiple entities. Please check!"
            )
        if len(complete_path) == 0:
            raise ValueError(f"The given path ({x}) points to no entity. Please check!")
        x = complete_path[0]
    return x


def parse_prune_ratio_vgg(sstr, num_layers=20):
    # example: [0-4:0.5, 5:0.6, 8-10:0.2]
    out = np.zeros(num_layers)
    if "[" in sstr:
        sstr = sstr.split("[")[1].split("]")[0]
    else:
        sstr = sstr.strip()
    for x in sstr.split(","):
        k = x.split(":")[0].strip()
        v = x.split(":")[1].strip()
        if k.isdigit():
            out[int(k)] = float(v)
        else:
            begin = int(k.split("-")[0].strip())
            end = int(k.split("-")[1].strip())
            out[begin : end + 1] = float(v)
    return list(out)


def kronecker(A, B):
    return torch.einsum("ab,cd->acbd", A, B).view(
        A.size(0) * B.size(0), A.size(1) * B.size(1)
    )


def np_to_torch(x):
    """
    np array to pytorch float tensor
    """
    x = np.array(x)
    x = torch.from_numpy(x).float()
    return x


def kd_loss(student_scores, teacher_scores, temp=1, weights=None):
    """Knowledge distillation loss: soft target"""
    p = F.log_softmax(student_scores / temp, dim=1)
    q = F.softmax(teacher_scores / temp, dim=1)
    # l_kl = F.kl_div(p, q, size_average=False) / student_scores.shape[0] # previous working loss
    if isinstance(weights, type(None)):
        l_kl = F.kl_div(
            p, q, reduction="batchmean"
        )  # 2020-06-21 @mst: Since 'size_average' is deprecated, use 'reduction' instead.
    else:
        l_kl = (F.kl_div(p, q, reduction="none").sum(dim=1) * weights).sum()
    return l_kl


def test(net, test_loader):
    n_example_test = 0
    total_correct = 0
    avg_loss = 0
    is_train = net.training
    net.eval()
    with torch.no_grad():
        pred_total = []
        label_total = []
        for _, (images, labels) in enumerate(test_loader):
            n_example_test += images.size(0)
            images = images.cuda()
            labels = labels.cuda()
            output = net(images)
            avg_loss += nn.CrossEntropyLoss()(output, labels).sum()
            pred = output.data.max(1)[1]
            total_correct += pred.eq(labels.data.view_as(pred)).sum()
            pred_total.extend(list(pred.data.cpu().numpy()))
            label_total.extend(list(labels.data.cpu().numpy()))

    acc = float(total_correct) / n_example_test
    avg_loss /= n_example_test

    # get accuracy per class
    n_class = output.size(1)
    acc_test = [0] * n_class
    cnt_test = [0] * n_class
    for p, l in zip(pred_total, label_total):
        acc_test[l] += int(p == l)
        cnt_test[l] += 1
    acc_per_class = []
    for c in range(n_class):
        acc_test[c] = 0 if cnt_test[c] == 0 else acc_test[c] / float(cnt_test[c])
        acc_per_class.append(acc_test[c])

    # return to the train state if necessary
    if is_train:
        net.train()
    return acc, avg_loss.item(), acc_per_class


def get_project_path(ExpID):
    full_path = glob.glob("Experiments/*%s*" % ExpID)
    assert (
        len(full_path) == 1
    )  # There should be only ONE folder with <ExpID> in its name.
    return full_path[0]


def parse_ExpID(path):
    """parse out the ExpID from 'path', which can be a file or directory.
    Example: Experiments/AE__ckpt_epoch_240.pth__LR1.5__originallabel__vgg13_SERVER138-20200829-202307/gen_img
    Example: Experiments/AE__ckpt_epoch_240.pth__LR1.5__originallabel__vgg13_SERVER-20200829-202307/gen_img
    """
    return "SERVER" + path.split("_SERVER")[1].split("/")[0]


def mkdirs(*paths, exist_ok=False):
    for p in paths:
        os.makedirs(p, exist_ok=exist_ok)


class EMA:
    """
    Exponential Moving Average for pytorch tensor
    """

    def __init__(self, mu):
        self.mu = mu
        self.history = {}

    def __call__(self, name, x):
        """
        Note: this func will modify x directly, no return value.
        x is supposed to be a pytorch tensor.
        """
        if self.mu > 0:
            assert 0 < self.mu < 1
            if name in self.history.keys():
                new_average = self.mu * self.history[name] + (1.0 - self.mu) * x.clone()
            else:
                new_average = x.clone()
            self.history[name] = new_average.clone()
            return new_average.clone()
        else:
            return x.clone()


# Exponential Moving Average
class EMA2:
    def __init__(self, mu):
        self.mu = mu
        self.shadow = {}

    def register(self, name, value):
        self.shadow[name] = value.clone()

    def __call__(self, name, x):
        assert name in self.shadow
        new_average = (1.0 - self.mu) * x + self.mu * self.shadow[name]
        self.shadow[name] = new_average.clone()
        return new_average


def register_ema(emas):
    for net, ema in emas:
        for name, param in net.named_parameters():
            if param.requires_grad:
                ema.register(name, param.data)


def apply_ema(emas):
    for net, ema in emas:
        for name, param in net.named_parameters():
            if param.requires_grad:
                param.data = ema(name, param.data)


colors = [
    "gray",
    "blue",
    "black",
    "yellow",
    "green",
    "yellowgreen",
    "gold",
    "royalblue",
    "peru",
    "purple",
]


def feat_visualize(ax, feat, label):
    """
    feat:  N x 2 # 2-d feature, N: number of examples
    label: N x 1
    """
    for ix in range(len(label)):
        x = feat[ix]
        y = label[ix]
        ax.scatter(x[0], x[1], color=colors[y], marker=".")
    return ax


def _remove_module_in_name(name):
    """remove 'module.' in the module name, caused by DataParallel, if any"""
    module_name_parts = name.split(".")
    module_name_parts_new = []
    for x in module_name_parts:
        if x != "module":
            module_name_parts_new.append(x)
    new_name = ".".join(module_name_parts_new)
    return new_name


def smart_weights_load(net, w_path, key=None, strict=True):
    r"""Load the weights of <w_path> into <net>."""
    common_weights_keys = [
        "T",
        "S",
        "G",
        "model",
        "state_dict",
        "state_dict_t",
    ]  # Used in previous projects, emprically set

    ckpt = torch.load(w_path, map_location=lambda storage, location: storage)

    # get state_dict
    if isinstance(ckpt, OrderedDict):
        state_dict = ckpt
    else:
        if key:
            state_dict = ckpt[key]
        else:
            intersection = [
                k
                for k in ckpt.keys()
                if k in common_weights_keys and isinstance(ckpt[k], OrderedDict)
            ]
            if len(intersection) == 1:
                k = intersection[0]
                state_dict = ckpt[k]
            else:
                print(
                    "Error: multiple or no model keys found in ckpt: %s. Please explicitly appoint one"
                    % intersection
                )
                exit(1)

    if (
        strict
    ):  # net and state_dict have exactly the same architecture (layer names etc. are exactly same)
        try:
            net.load_state_dict(state_dict)
        except:
            ckpt_data_parallel = False
            for k, v in state_dict.items():
                if k.startswith("module."):
                    ckpt_data_parallel = True  # DataParallel was used in the ckpt
                    break

            if ckpt_data_parallel:
                # If ckpt used DataParallel, then the reason of the load failure above should be that the <net> does not use
                # DataParallel. Therefore, remove the surfix 'module.' in ckpt.
                new_state_dict = OrderedDict()
                for k, v in state_dict.items():
                    param_name = k.split("module.")[-1]
                    new_state_dict[param_name] = v
            else:
                # Similarly, if ckpt didn't use DataParallel, here we add the surfix 'module.'.
                new_state_dict = OrderedDict()
                for k, v in state_dict.items():
                    param_name = "module." + k
                    new_state_dict[param_name] = v
            net.load_state_dict(new_state_dict)

    else:
        # Here is the case that <net> and ckpt only have part of weights in common. Then load them by module name:
        # for every named module in <net>, if ckpt has a module of the same (or contextually similar) name, then they are matched and weights are loaded from ckpt to <net>.
        for name, m in net.named_modules():
            print(name)

        for name, m in net.named_modules():
            if name:
                print('loading weights for module "%s" in the network' % name)
                new_name = _remove_module_in_name(name)

                # find the matched module name
                matched_param_name = ""
                for k in ckpt.keys():
                    new_k = _remove_module_in_name(k)
                    if new_name == new_k:
                        matched_param_name = k
                        break

                # load weights
                if matched_param_name:
                    m.weight.copy_(ckpt[matched_param_name])
                    print(
                        "net module name: '%s' <- '%s' (ckpt module name)"
                        % (name, matched_param_name)
                    )
                else:
                    print(
                        "Error: cannot find matched module in ckpt for module '%s' in net. Please check manually."
                        % name
                    )
                    exit(1)


# parse wanted value from accuracy print log
def parse_acc_log(line, key, type_func=float):
    line_seg = line.strip().lower().split()
    for i in range(len(line_seg)):
        if key in line_seg[i]:
            break
    if i == len(line_seg) - 1:
        return None  # did not find the <key> in this line
    try:
        value = type_func(line_seg[i + 1])
    except:
        value = type_func(line_seg[i + 2])
    return value


def get_layer_by_index(net, index):
    cnt = -1
    for _, m in net.named_modules():
        if isinstance(m, nn.Conv2d) or isinstance(m, nn.Linear):
            cnt += 1
            if cnt == index:
                return m
    return None


def get_total_index_by_learnable_index(net, learnable_index):
    """
    learnable_index: index when only counting learnable learnable_layers (conv or fc, no bn);
    total_index: count relu, pooling etc in.
    """
    layer_type_considered = [
        nn.Conv2d,
        nn.ReLU,
        nn.LeakyReLU,
        nn.PReLU,
        nn.BatchNorm2d,
        nn.MaxPool2d,
        nn.AvgPool2d,
        nn.Linear,
    ]
    cnt_total = -1
    cnt_learnable = -1
    for _, m in net.named_modules():
        cond = [isinstance(m, x) for x in layer_type_considered]
        if any(cond):
            cnt_total += 1
            if isinstance(m, nn.Conv2d) or isinstance(m, nn.Linear):
                cnt_learnable += 1
                if cnt_learnable == learnable_index:
                    return cnt_total
    return None


def cal_correlation(x, coef=False):
    """Calculate the correlation matrix for a pytorch tensor.
    Input shape: [n_sample, n_attr]
    Output shape: [n_attr, n_attr]
    Refer to: https://github.com/pytorch/pytorch/issues/1254
    """
    # calculate covariance matrix
    y = x - x.mean(dim=0)
    c = y.t().mm(y) / (y.size(0) - 1)

    if coef:
        # normalize covariance matrix
        d = torch.diag(c)
        stddev = torch.pow(d, 0.5)
        c = c.div(stddev.expand_as(c))
        c = c.div(stddev.expand_as(c).t())

        # clamp between -1 and 1
        # probably not necessary but numpy does it
        c = torch.clamp(c, -1.0, 1.0)
    return c


def get_class_corr(loader, model):
    model.eval().cuda()
    logits = 0
    n_batch = len(loader)
    with torch.no_grad():
        for ix, data in enumerate(loader):
            input = data[0]
            print("[%d/%d] -- forwarding" % (ix, n_batch))
            input = input.float().cuda()
            if type(logits) == int:
                logits = model(input)  # [batch_size, n_class]
            else:
                logits = torch.cat([logits, model(input)], dim=0)
    # Use numpy:
    # logits -= logits.mean(dim=0)
    # logits = logits.data.cpu().numpy()
    # corr = np.corrcoef(logits, rowvar=False)

    # Use pytorch
    corr = cal_correlation(logits, coef=True)
    return corr


def cal_acc(logits, y):
    pred = logits.argmax(dim=1)
    acc = pred.eq(y.data.view_as(pred)).sum().float() / y.size(0)
    return acc


class Timer:
    """Log down iteration time and predict the left time for the left iterations"""

    def __init__(self, total_epoch):
        self.total_epoch = total_epoch
        self.time_stamp = []

    def predict_finish_time(self, ave_window=3):
        self.time_stamp.append(time.time())  # update time stamp
        if len(self.time_stamp) == 1:
            return "only one time stamp, not enough to predict"
        interval = []
        for i in range(len(self.time_stamp) - 1):
            t = self.time_stamp[i + 1] - self.time_stamp[i]
            interval.append(t)
        sec_per_epoch = np.mean(interval[-ave_window:])
        left_t = sec_per_epoch * (self.total_epoch - len(interval))
        finish_t = left_t + time.time()
        finish_t = time.strftime("%Y/%m/%d-%H:%M", time.localtime(finish_t))
        total_t = "%.2fh" % ((np.sum(interval) + left_t) / 3600.0)
        return finish_t + " (speed: %.2fs per timing, total_time: %s)" % (
            sec_per_epoch,
            total_t,
        )

    def __call__(self):
        return self.predict_finish_time()


class Dataset_npy_batch(Dataset):
    def __init__(self, npy_dir, transform, f="batch.npy"):
        self.data = np.load(os.path.join(npy_dir, f), allow_pickle=True)
        self.transform = transform

    def __getitem__(self, index):
        img = Image.fromarray(self.data[index][0])
        img = self.transform(img)
        label = self.data[index][1]
        label = torch.LongTensor([label])[0]
        return img.squeeze(0), label

    def __len__(self):
        return len(self.data)


class Dataset_lmdb_batch(Dataset):
    """Dataset to load a lmdb data file."""

    def __init__(self, lmdb_path, transform):
        import lmdb

        env = lmdb.open(lmdb_path, readonly=True)
        with env.begin() as txn:
            self.data = [value for key, value in txn.cursor()]
        self.transform = transform

    def __getitem__(self, index):
        img, label = pickle.loads(self.data[index])  # PIL image
        if self.transform:
            img = self.transform(img)
        return img, label

    def __len__(self):
        return len(self.data)


class AccuracyManager:
    def __init__(self):
        import pandas as pd

        self.accuracy = pd.DataFrame()

    def update(self, time, acc1, acc5=None):
        acc = pd.DataFrame(
            [[time, acc1, acc5]], columns=["time", "acc1", "acc5"]
        )  # time can be epoch or step
        self.accuracy = self.accuracy.append(acc, ignore_index=True)

    def get_best_acc(self, criterion="acc1"):
        assert criterion in ["acc1", "acc5"]
        acc = self.accuracy.sort_values(by=criterion)  # ascending sort
        best = acc.iloc[-1]  # the last row
        time, acc1, acc5 = best.time, best.acc1, best.acc5
        return time, acc1, acc5

    def get_last_acc(self):
        last = self.accuracy.iloc[-1]
        time, acc1, acc5 = last.time, last.acc1, last.acc5
        return time, acc1, acc5


def format_acc_log(acc1_set, lr, acc5=None, time_unit="Epoch"):
    """return uniform format for the accuracy print"""
    acc1, acc1_time, acc1_best, acc1_best_time = acc1_set
    if acc5:
        line = "Acc1 %.4f Acc5 %.4f @ %s %d (Best_Acc1 %.4f @ %s %d) LR %s" % (
            acc1,
            acc5,
            time_unit,
            acc1_time,
            acc1_best,
            time_unit,
            acc1_best_time,
            lr,
        )
    else:
        line = "Acc1 %.4f @ %s %d (Best_Acc1 %.4f @ %s %d) LR %s" % (
            acc1,
            time_unit,
            acc1_time,
            acc1_best,
            time_unit,
            acc1_best_time,
            lr,
        )
    return line


def get_lambda(alpha=1.0):
    """Return lambda"""
    if alpha > 0.0:
        lam = np.random.beta(alpha, alpha)
    else:
        lam = 1.0
    return lam


# refer to: 2018-ICLR-mixup
# https://github.com/facebookresearch/mixup-cifar10/blob/eaff31ab397a90fbc0a4aac71fb5311144b3608b/train.py#L119
def mixup_data(x, y, alpha=1.0, use_cuda=True):
    """Returns mixed inputs, pairs of targets, and lambda"""
    if alpha > 0:
        lam = np.random.beta(alpha, alpha)
    else:
        lam = 1

    batch_size = x.size()[0]
    if use_cuda:
        index = torch.randperm(batch_size).cuda()
    else:
        index = torch.randperm(batch_size)

    mixed_x = lam * x + (1 - lam) * x[index, :]
    y_a, y_b = y, y[index]
    return mixed_x, y_a, y_b, lam


def mixup_criterion(criterion, pred, y_a, y_b, lam):
    return lam * criterion(pred, y_a) + (1 - lam) * criterion(pred, y_b)


def visualize_filter(
    layer,
    layer_id,
    save_dir,
    n_filter_plot=16,
    n_channel_plot=16,
    pick_mode="rand",
    plot_abs=True,
    prefix="",
    ext=".pdf",
):
    """layer is a pytorch model layer"""
    w = layer.weight.data.cpu().numpy()  # shape: [N, C, H, W]
    if plot_abs:
        w = np.abs(w)
    n, c = w.shape[0], w.shape[1]
    n_filter_plot = min(n_filter_plot, n)
    n_channel_plot = min(n_channel_plot, c)
    if pick_mode == "rand":
        filter_ix = np.random.permutation(n)[:n_filter_plot]  # filter indexes to plot
        channel_ix = np.random.permutation(c)[
            :n_channel_plot
        ]  # channel indexes to plot
    else:
        filter_ix = list(range(n_filter_plot))
        channel_ix = list(range(n_channel_plot))

    # iteration for plotting
    for i in filter_ix:
        f_avg = np.mean(w[i], axis=0)
        fig, ax = plt.subplots()
        im = ax.imshow(f_avg, cmap="jet")
        # make a beautiful colorbar
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size=0.05, pad=0.05)
        fig.colorbar(im, cax=cax, orientation="vertical")
        save_path = (
            "%s/filter_visualize__%s__layer%s__filter%s__average_cross_channel"
            % (save_dir, prefix, layer_id, i)
        )  # prefix is usually a net name
        fig.savefig(save_path + ext, bbox_inches="tight")
        plt.close(fig)

        for j in channel_ix:
            f = w[i][j]
            fig, ax = plt.subplots()
            im = ax.imshow(f, cmap="jet")
            # make a beautiful colorbar
            divider = make_axes_locatable(ax)
            cax = divider.append_axes("right", size=0.05, pad=0.05)
            fig.colorbar(im, cax=cax, orientation="vertical")
            save_path = "%s/filter_visualize__%s__layer%s__filter%s__channel%s" % (
                save_dir,
                prefix,
                layer_id,
                i,
                j,
            )
            fig.savefig(save_path + ext, bbox_inches="tight")
            plt.close(fig)


def visualize_feature_map(
    fm,
    layer_id,
    save_dir,
    n_channel_plot=16,
    pick_mode="rand",
    plot_abs=True,
    prefix="",
    ext=".pdf",
):
    fm = fm.clone().detach()
    fm = fm.data.cpu().numpy()[
        0
    ]  # shape: [N, C, H, W], N is batch size. Default: batch size should be 1
    if plot_abs:
        fm = np.abs(fm)
    c = fm.shape[0]
    n_channel_plot = min(n_channel_plot, c)
    if pick_mode == "rand":
        channel_ix = np.random.permutation(c)[
            :n_channel_plot
        ]  # channel indexes to plot
    else:
        channel_ix = list(range(n_channel_plot))

    # iteration for plotting
    fm_avg = np.mean(fm, axis=0)
    fig, ax = plt.subplots()
    im = ax.imshow(fm_avg, cmap="jet")
    # make a beautiful colorbar
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size=0.05, pad=0.05)
    fig.colorbar(im, cax=cax, orientation="vertical")
    save_path = "%s/featmap_visualization__%s__layer%s__average_cross_channel" % (
        save_dir,
        prefix,
        layer_id,
    )  # prefix is usually a net name
    fig.savefig(save_path + ext, bbox_inches="tight")
    plt.close(fig)

    for j in channel_ix:
        f = fm[j]
        fig, ax = plt.subplots()
        im = ax.imshow(f, cmap="jet")
        # make a beautiful colorbar
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size=0.05, pad=0.05)
        fig.colorbar(im, cax=cax, orientation="vertical")
        save_path = "%s/featmap_visualization__%s__layer%s__channel%s" % (
            save_dir,
            prefix,
            layer_id,
            j,
        )
        fig.savefig(save_path + ext, bbox_inches="tight")
        plt.close(fig)


def add_noise_to_model(model, std=0.01):
    model = copy.deepcopy(model)  # do not modify the original model
    for name, module in model.named_modules():
        if isinstance(
            module, (nn.Conv2d, nn.Linear, nn.BatchNorm2d)
        ):  # all learnable params for a typical DNN
            w = module.weight
            w.data += torch.randn_like(w) * std
    return model


# pt1.9 does not have module 'zero_gradients' in 'torch.autograd.gradcheck', so implement it here, referring to:
# https://github.com/pytorch/pytorch/blob/819d4b2b83fa632bf65d14f6af80a09e7476e87e/torch/autograd/gradcheck.py#L15
def iter_gradients(x):
    if isinstance(x, Variable):
        if x.requires_grad and x.grad is not None:
            yield x.grad.data
    else:
        for elem in x:
            for result in iter_gradients(elem):
                yield result


def zero_gradients(i):
    for t in iter_gradients(i):
        t.zero_()


# Refer to: https://github.com/ast0414/adversarial-example/blob/26ee4144a1771d3a565285e0a631056a6f42d49c/craft.py#L6
def compute_jacobian(inputs, output):
    """
    :param inputs: Batch X Size (e.g. Depth X Width X Height)
    :param output: Batch X Classes
    :return: jacobian: Batch X Classes X Size
    """
    # from torch.autograd.gradcheck import zero_gradients # cannot be imported for pt1.9
    assert inputs.requires_grad
    num_classes = output.size()[1]

    jacobian = torch.zeros(num_classes, *inputs.size())
    grad_output = torch.zeros(*output.size())
    if inputs.is_cuda:
        grad_output = grad_output.cuda()
        jacobian = jacobian.cuda()

    for i in range(num_classes):
        zero_gradients(inputs)
        grad_output.zero_()
        grad_output[:, i] = 1
        output.backward(grad_output, retain_graph=True)
        jacobian[i] = inputs.grad.data

    return torch.transpose(jacobian, dim0=0, dim1=1)


def get_jacobian_singular_values(
    model, data_loader, num_classes, n_loop=20, print_func=print, rand_data=False
):
    jsv, jsv_diff, condition_number = [], [], []
    if rand_data:
        picked_batch = np.random.permutation(len(data_loader))[:n_loop]
    else:
        picked_batch = list(range(n_loop))
    for i, (images, target) in enumerate(data_loader):
        if i in picked_batch:
            images, target = images.cuda(), target.cuda()
            batch_size = images.size(0)
            images.requires_grad = True  # for Jacobian computation
            output = model(images)
            jacobian = compute_jacobian(
                images, output
            )  # shape [batch_size, num_classes, num_channels, input_width, input_height]
            jacobian = jacobian.view(
                batch_size, num_classes, -1
            )  # shape [batch_size, num_classes, num_channels*input_width*input_height]
            u, s, v = torch.svd(
                jacobian
            )  # u: [batch_size, num_channels*input_width*input_height, num_classes], s: [batch_size, num_classes], v: [batch_size, num_channels*input_width*input_height, num_classes]
            s = s.data.cpu().numpy()
            jsv.append(s)
            jsv_diff.append((s - 1) ** 2)
            condition_number.append(s.max(axis=1) / s.min(axis=1))
            print_func("[%3d/%3d] calculating Jacobian..." % (i, len(data_loader)))
    jsv = np.concatenate(jsv)
    condition_number = np.concatenate(condition_number)
    return jsv, jsv_diff, condition_number


def approximate_entropy(X, num_bins=10, esp=1e-30):
    """X shape: [num_sample, n_var], numpy array."""
    entropy = []
    for di in range(X.shape[1]):
        samples = X[:, di]
        bins = np.linspace(samples.min(), samples.max(), num=num_bins + 1)
        prob = np.histogram(samples, bins=bins, density=False)[0] / len(samples)
        entropy.append(
            (-np.log2(prob + esp) * prob).sum()
        )  # esp for numerical stability when prob = 0
    return np.mean(entropy)


# matplotlib utility functions
def set_ax(ax):
    """This will modify ax in place."""
    # set background
    ax.grid(color="white")
    ax.set_facecolor("whitesmoke")

    # remove axis line
    ax.spines["right"].set_visible(False)
    ax.spines["left"].set_visible(False)
    ax.spines["top"].set_visible(False)
    ax.spines["bottom"].set_visible(False)

    # remove tick but keep the values
    ax.xaxis.set_ticks_position("none")
    ax.yaxis.set_ticks_position("none")


def parse_value(line, key, type_func=float, exact_key=True):
    """Parse a line with the key"""
    try:
        if exact_key:  # back compatibility
            value = line.split(key)[1].strip().split()[0]
            if value.endswith(")"):  # hand-fix case: "Epoch 23)"
                value = value[:-1]
            value = type_func(value)
        else:
            line_seg = line.split()
            for i in range(len(line_seg)):
                if key in line_seg[i]:  # example: 'Acc1: 0.7'
                    break
            if i == len(line_seg) - 1:
                return None  # did not find the <key> in this line
            value = type_func(line_seg[i + 1])
        return value
    except:
        print('Got error for line: "%s". Please check.' % line)


def to_tensor(x):
    x = np.array(x)
    x = torch.from_numpy(x).float()
    return x


def denormalize_image(x, mean, std):
    """x shape: [N, C, H, W], batch image"""
    x = x.cuda()
    mean = to_tensor(mean).cuda()
    std = to_tensor(std).cuda()
    mean = mean.unsqueeze(0).unsqueeze(2).unsqueeze(3)  # shape: [1, C, 1, 1]
    std = std.unsqueeze(0).unsqueeze(2).unsqueeze(3)
    x = std * x + mean
    return x


def normalize_image(x, mean, std):
    """x shape: [N, C, H, W], batch image"""
    x = x.cuda()
    mean = to_tensor(mean).cuda()
    std = to_tensor(std).cuda()
    mean = mean.unsqueeze(0).unsqueeze(2).unsqueeze(3)  # shape: [1, C, 1, 1]
    std = std.unsqueeze(0).unsqueeze(2).unsqueeze(3)
    x = (x - mean) / std
    return x


def make_one_hot(labels, C):  # labels: [N]
    """turn a batch of labels to the one-hot form"""
    labels = labels.unsqueeze(1)  # [N, 1]
    one_hot = torch.zeros(labels.size(0), C).cuda()
    target = one_hot.scatter_(1, labels, 1)
    return target


class AverageMeter(object):
    """Computes and stores the average and current value"""

    def __init__(self, name, fmt=":f"):
        self.name = name
        self.fmt = fmt
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

    def __str__(self):
        fmtstr = "{name} {val" + self.fmt + "} ({avg" + self.fmt + "})"
        return fmtstr.format(**self.__dict__)


class ProgressMeter(object):
    def __init__(self, num_batches, meters, prefix=""):
        self.batch_fmtstr = self._get_batch_fmtstr(num_batches)
        self.meters = meters
        self.prefix = prefix

    def display(self, batch):
        entries = [self.prefix + self.batch_fmtstr.format(batch)]
        entries += [str(meter) for meter in self.meters]
        print("\t".join(entries))

    def _get_batch_fmtstr(self, num_batches):
        num_digits = len(str(num_batches // 1))
        fmt = "{:" + str(num_digits) + "d}"
        return "[" + fmt + "/" + fmt.format(num_batches) + "]"


def adjust_learning_rate(optimizer, epoch, args):
    """Sets the learning rate to the initial LR decayed by 10 every 30 epochs"""
    lr = args.lr * (0.1 ** (epoch // 30))
    for param_group in optimizer.param_groups:
        param_group["lr"] = lr
    return lr


def accuracy(output, target, topk=(1,)):
    """Computes the accuracy over the k top predictions for the specified values of k"""
    with torch.no_grad():
        maxk = max(topk)
        batch_size = target.size(0)
        _, pred = output.topk(maxk, 1, True, True)
        pred = pred.t()  # shape [maxk, batch_size]
        correct = pred.eq(
            target.view(1, -1).expand_as(pred)
        )  # target shape: [batch_size] -> [1, batch_size] -> [maxk, batch_size]
        res = []
        for k in topk:
            # correct_k = correct[:k].view(-1).float().sum(0, keepdim=True) # Because of pytorch new versions, this does not work anymore (pt1.3 is okay, pt1.9 not okay).
            correct_k = correct[:k].flatten().float().sum(0, keepdim=True)
            res.append(correct_k.mul_(100.0 / batch_size))
        return res


class LossLine:
    """Format loss items for easy print."""

    def __init__(self):
        self.log_dict = OrderedDict()
        self.formats = OrderedDict()

    def update(self, key, value, format):
        self.log_dict[key] = value
        self.formats[key] = format

    def format(self, sep=" "):
        out = []
        for k, v in self.log_dict.items():
            item = f"{k} {v:{self.formats[k]}}"
            out.append(item)
        return sep.join(out)


class EmptyClass:
    pass


def update_args(args):
    """Update arguments of configargparse"""
    arg_dict = copy.deepcopy(args.__dict__)
    for k, v in arg_dict.items():
        if "." in k:  # TODO-@mst: hardcode pattern, may be risky
            module, arg = k.split(".")  # e.g., "deepmixup.depth"
            if arg_dict[f"{module}.ON"]:  # this module is being used
                if not hasattr(args, module):
                    args.__setattr__(module, EmptyClass())  # set to a blank class
                args.__dict__[module].__dict__[
                    arg
                ] = v  # args.'deepmixup.depth' = 10 --> args.deepmixup.depth = 10
            args.__delattr__(k)
    return args


def check_kernel_spatial_dist(model):
    for name, module in model.named_modules():
        if isinstance(module, (nn.Conv2d, nn.Linear)):
            spatial = module.weight.data.abs().mean(dim=(0, 1))
            print(f"{name} kernel spatial dist:")
            print(spatial.data.cpu().numpy())


def print_format(x, fmt="%.3f", sep=" "):
    return sep.join([fmt % xi for xi in x])


def check_grad_norm(model):
    for name, module in model.named_modules():
        if isinstance(module, (nn.Conv2d, nn.Linear, nn.BatchNorm2d)):
            grad_abs = module.weight.grad.abs().view(module.weight.shape[0], -1)
            grad_abs_mean = grad_abs.mean(dim=-1)
            logstr = print_format(grad_abs_mean)
            print(f"[{name:>20s}] layer grad norm: {logstr}")


def check_grad_stats(model):
    for name, module in model.named_modules():
        if isinstance(module, (nn.Conv2d, nn.Linear, nn.BatchNorm2d)):
            grad = module.weight.grad
            grad_abs = grad.abs()
            grad_abs_var, grad_abs_mean = grad_abs.var(), grad_abs.mean()
            print(
                f"[{name:>20s}] layer grad_abs: mean {grad_abs_mean:.8f} variance {grad_abs_var:.8f}"
            )


def check_grad_stats_v2(model, layers, selected_layers, grad_stats, is_print):
    modules = {}
    for n, m in model.named_modules():
        modules[n] = m
    ordered_modules = OrderedDict()
    for n in layers:
        ordered_modules[n] = modules[n]
    del modules
    for n, m in ordered_modules.items():
        if n in selected_layers and hasattr(m, "weight") and m.weight.grad is not None:
            grad_abs = m.weight.grad.abs().cpu().data.numpy()
            grad_abs_var, grad_abs_mean = grad_abs.var(), grad_abs.mean()
            if is_print:
                print(
                    f"{layers[n].print_prefix} layer grad_abs: mean {grad_abs_mean:.8f} variance {grad_abs_var:.8f}"
                )
            if n not in grad_stats:
                grad_stats[n] = []
            grad_stats[n].append([grad_abs_mean, grad_abs_var])


def check_weight_stats(model):
    for name, module in model.named_modules():
        if isinstance(module, (nn.Conv2d, nn.Linear, nn.BatchNorm2d)):
            w_mean, w_std = torch.mean(module.weight), torch.std(module.weight)
            print(f"[{name:>20s}] weight mean: {w_mean:>7.4f} std: {w_std:>7.4f}")


def update_args_from_file(args, config_path):
    import json, yaml

    with open(config_path) as f:
        if config_path.endswith(".json"):
            params = json.load(f)
        elif config_path.endswith(".yaml"):
            params = yaml.load(f, Loader=yaml.FullLoader)
        else:
            raise NotImplementedError
    for k, v in params.items():
        assert k in args.__dict__, f"'{k}'' is not in original args. Please check!"
        args.__dict__[k] = v
    return args


def apply_mask(model, mask, forward=True):
    r"""Apply mask to model.
    Args:
        model (PyTorch model)
        mask (dict): key is module name, value is Tensor
    """
    for name, m in model.named_modules():
        if name in mask:
            if forward:
                m.weight.data.mul_(mask[name])
            else:  # Backward, masking gradients. Not checked this feature, may not work!
                m.weight.grad.data.mul_(mask[name])
    return model


def isfloat(num):
    r"""Check if a number is float."""
    try:
        float(num)
        return True
    except ValueError:
        return False


def moving_average(x, N=10):
    r"""Refer to: https://stackoverflow.com/questions/13728392/moving-average-or-running-mean"""
    import scipy.ndimage as ndi

    return ndi.uniform_filter1d(x, N, mode="constant", origin=-(N // 2))[: -(N - 1)]


smooth = moving_average  # To maintain back-compatibility


def get_exp_name_id(exp_path):
    r"""arg example: Experiments/kd-vgg13vgg8-cifar100-Temp40_SERVER5-20200727-220318
    or kd-vgg13vgg8-cifar100-Temp40_SERVER5-20200727-220318
    or Experiments/kd-vgg13vgg8-cifar100-Temp40_SERVER5-20200727-220318/weights/ckpt.pth
    """
    seps = exp_path.split(os.sep)
    for s in seps:
        if "_SERVER" in s:
            exp_id = s.split("-")[-1]
            assert exp_id.isdigit()
            ExpID = "SERVER" + s.split("_SERVER")[1]
            exp_name = s.split("_SERVER")[0]
            date = s.split("-")[-2]
            return ExpID, exp_id, exp_name, date


def get_script_from_log(log_path, max_lines=10):
    r"""Get the script from a log txt file."""
    cnt = 0
    for line in open(log_path):
        cnt += 1
        line = line.strip()
        if line.startswith("CUDA_VISIBLE_DEVICES=") or line.startswith("python"):
            return line
        if cnt == max_lines:
            return None


def run_shell_command(cmd, inarg=None):
    r"""Run shell command and return the output (string) in a list"""
    cmd = " ".join(cmd.split())
    if " | " in cmd:  # Refer to: https://stackoverflow.com/a/13332300/12554945
        cmds = cmd.split(" | ")
        assert len(cmds) == 2, "Only support one pipe now"
        fn = subprocess.Popen(cmds[0].split(), stdout=subprocess.PIPE)
        result = subprocess.run(
            cmds[1].split(), stdin=fn.stdout, stdout=subprocess.PIPE
        )
    else:
        result = subprocess.run(cmd.split(), stdout=subprocess.PIPE)
    return result.stdout.decode("utf-8").strip().split("\n")


def print_runtime(fn):
    r"""Print the running time of a routine."""

    @functools.wraps(fn)
    def wrapper(*args, **kw):
        t0 = time.time()
        ret = fn(*args, **kw)
        t1 = time.time()
        print(f'( "{fn.__name__}" executed in {t1 - t0:.4f}s )')
        return ret

    return wrapper


def get_arg(args, key):
    return args.__dict__.get(key)


def scp_experiment(scp_script, logger, args, mv=False):
    userip = logger.userip
    ExpID = logger.ExpID
    experiments_dir = args.experiments_dir
    exp_name = (
        args.experiment_name if hasattr(args, "experiment_name") else args.project_name
    )

    lines = open(scp_script).readlines()
    for line in lines:
        if " scp " in line and "@" in line:
            break
    for i in line.strip().split():
        if "@" in i and ":" in i:
            hub_userip = i.split(":")[0]
            break
    need_scp = not args.debug and userip != hub_userip
    if need_scp:
        script = f"sh {scp_script} {experiments_dir} {exp_name}_{ExpID}"
        os.system(script)
        if mv:
            if not os.path.exists(f"{experiments_dir}/Trash"):
                os.makedirs(f"{experiments_dir}/Trash")
            os.system(
                f"mv {experiments_dir}/{exp_name}_{ExpID} {experiments_dir}/Trash"
            )
    return need_scp


def scp_experiment_v2(
    logger, args, scp_script="scripts/scp_experiments_to_hub.sh", mv=False, init=False
):
    userip = logger.userip
    ExpID = logger.ExpID
    experiments_dir = args.experiments_dir
    exp_name = (
        args.experiment_name if hasattr(args, "experiment_name") else args.project_name
    )

    lines = open(scp_script).readlines()
    for line in lines:
        if " scp " in line and "@" in line:
            break
    for i in line.strip().split():
        if "@" in i and ":" in i:
            hub_userip = i.split(":")[0]
            break
    need_scp = not args.debug and userip != hub_userip
    word = "Initial" if init else "Final"
    if need_scp:
        print(f"{word} scp experiments to hub...", end="")
        t0 = time.time()
        script = f"sh {scp_script} {experiments_dir} {exp_name}_{ExpID}"
        os.system(script)
        print(f", done! Time: {time.time() - t0:.1f}s", unprefix=True)
        if mv:
            if not os.path.exists(f"{experiments_dir}/Trash"):
                os.makedirs(f"{experiments_dir}/Trash")
            os.system(
                f"mv {experiments_dir}/{exp_name}_{ExpID} {experiments_dir}/Trash"
            )



from distutils.core import setup

setup(
    name="smilelogging",  # How you named your package folder (MyLib)
    packages=["smilelogging"],  # Chose the same as "name"
    version="0.6",  # Start with a small number and increase it with every change you make
    license="MIT",  # Chose a license from here: https://help.github.com/articles/licensing-a-repository
    description="Python logging package for easy reproducible experimenting in research.",  # Give a short description about your library
    author="Huan Wang",  # Type in your name
    author_email="wanghuan@westlake.edu.cn",  # Type in your E-Mail
    url="https://github.com/MingSun-Tse/smilelogging",  # Provide either the link to your github or to your website
    download_url="https://github.com/MingSun-Tse/smilelogging/archive/refs/tags/v0.6.tar.gz",  # I explain this later on
    keywords=[
        "logging",
        "reproducible experimenting",
    ],  # Keywords that define your package best
    install_requires=[  # I get to this in a second
        "numpy",
        "pandas",
        "matplotlib",
        "Pillow",
        "PyYAML",
        "lmdb",
        "configargparse",
        "pytz",
        "colorama",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",  # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        "Intended Audience :: Science/Research",  # Define that your audience are developers
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",  # Again, pick a license
        "Programming Language :: Python :: 3",  # Specify which Python versions that you want to support
    ],
)



import builtins
import getpass
import glob
import os
import socket
import subprocess
import sys
import time
import traceback
from collections import OrderedDict
from datetime import datetime
from fnmatch import fnmatch
import shutil
from types import SimpleNamespace

import numpy as np
import yaml
import warnings
from glob import glob

from smilelogging.slutils import bold, blue, green, red, yellow

pjoin = os.path.join
TIMEZONE = datetime.now().astimezone().tzinfo  # Use local time zone.


class Folder:
    def __init__(self, path: str):
        self.path = path
        self._subfolders = {}

    def __repr__(self):
        return f"Folder({self.path})"

    def add_subfolder(self, key, subfolder):
        self._subfolders[key] = subfolder

    def __getattr__(self, name):
        if name in self._subfolders:
            return self._subfolders[name]
        raise AttributeError(f"'{self.name}' object has no attribute '{name}'")


def print_tree(directory, indent="", is_root=True):
    """Print a directory in a tree-like fashion."""
    # Print the root directory name if it's the root call
    if is_root:
        print(directory)

    # Get list of all files and directories in the given directory
    items = os.listdir(directory)

    # Iterate over each item
    for index, item in enumerate(items):
        # Create the full path for the item
        item_path = os.path.join(directory, item)

        # Check if it's the last item in the list for formatting
        is_last = index == len(items) - 1

        # Print the current item with appropriate indentation
        print(indent + ("└── " if is_last else "├── ") + item)

        # If it's a directory, recurse into it
        if os.path.isdir(item_path):
            print_tree(
                item_path, indent + ("    " if is_last else "│   "), is_root=False
            )


def _pretty_dict_format(a_dict: dict):
    # Build a key map for later sorting.
    key_map = {}
    for k in a_dict:
        k_lower = k.lower()
        if k_lower in key_map:
            key_map[k_lower + "_" + k_lower] = k
        else:
            key_map[k_lower] = k

    # Print in the order of sorted lower keys.
    logstr = ""
    for k_ in sorted(key_map.keys()):
        real_key = key_map[k_]
        logstr += "('%s': %s) " % (real_key, a_dict[real_key])
    return logstr


def create_folder_structure(
    experiments_path: str,
    experiment_folder: str,
    experiment_folder_structure: dict,
    rank: int,
):
    """Create folders based on provided folder structure.
    Args:
        experiments_path: Path of the folder that hosts all experiments.
        experiment_folder: Name of the experiment folder to be created.
        experiment_folder_structure: A dict to indicate the folder structure.
        rank: Integer. The global rank of current process.

    Returns:
        expfolder: Experiment folder, which is an instance of class Folder.
    """
    uniform_experiment_name = "experiment"
    expfolder = Folder(path=experiments_path)
    if rank > 0:
        expfolder = Folder(path=pjoin(experiments_path, "OtherRanks"))

    def _create_structure(base_folder, structure):
        base_folder_path = base_folder.path
        for subfolder, content in structure.items():
            key = subfolder
            if subfolder.startswith("<") and subfolder.endswith(
                ">"
            ):  # This indicates the name should be replaced. E.g., <experiment_folder>
                subfolder = experiment_folder
                key = uniform_experiment_name

            # Create the full path for the current item
            subfolder_path = os.path.join(base_folder_path, subfolder)
            subfolder = Folder(path=subfolder_path)
            os.makedirs(
                subfolder_path, mode=0o777, exist_ok=True
            )  # Use mode 777 for docker. Otherwise, the folder will belong to root.
            base_folder.add_subfolder(key, subfolder)

            # Process the content.
            if content is not None:
                if isinstance(content, dict):
                    # Recursively create sub-structure
                    _create_structure(subfolder, content)
                else:
                    content = content.split()  # The content may be multiple files.
                    for each_content in content:
                        each_content = each_content.strip()
                        with open(pjoin(subfolder_path, each_content), "w") as f:
                            f.write("")

    # Start creating from the root of the structure
    _create_structure(expfolder, experiment_folder_structure)
    return expfolder


def run_shell_command(cmd):
    """Run shell command and return the output (string) in a list.
    Args:
        cmd: shell command

    Returns:
        result: a list of the output returned by that shell command
    """
    cmd = " ".join(cmd.split())
    if " | " in cmd:  # Refer to: https://stackoverflow.com/a/13332300/12554945
        cmds = cmd.split(" | ")
        assert len(cmds) == 2, "Only support one pipe now"
        fn = subprocess.Popen(cmds[0].split(), stdout=subprocess.PIPE)
        result = subprocess.run(
            cmds[1].split(), stdin=fn.stdout, stdout=subprocess.PIPE
        )
    else:
        result = subprocess.run(cmd.split(), stdout=subprocess.PIPE)
    result = result.stdout.decode("utf-8").strip().split("\n")
    return result


def check_command_installed(cmd: str):
    """Check whether a system command, indicated by `cmd`, is installed."""
    try:
        result = subprocess.run(
            [f"{cmd}"], stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        return result.returncode == 0
    except FileNotFoundError:
        return False


def moving_average(x, N=10):
    """Refer to: https://stackoverflow.com/questions/13728392/moving-average-or-running-mean"""
    import scipy.ndimage as ndi

    return ndi.uniform_filter1d(x, N, mode="constant", origin=-(N // 2))[: -(N - 1)]


def get_experiment_path(ExpID):
    full_path = glob("Experiments/*%s*" % ExpID)
    assert (
        len(full_path) == 1
    ), "There should be only ONE folder with <ExpID> in its name"
    return full_path[0]


def parse_ExpID(path):
    """parse out the ExpID from 'path', which can be a file or directory.
    Example: Experiments/AE__ckpt_epoch_240.pth__LR1.5__originallabel__vgg13_SERVER138-20200829-202307/gen_img
    Example: Experiments/AE__ckpt_epoch_240.pth__LR1.5__originallabel__vgg13_SERVER-20200829-202307/gen_img
    """
    rank = ""
    if "RANK" in path:
        rank = path.split("RANK")[1].split("-")[0]
        rank = f"RANK{rank}-"
    return rank + "SERVER" + path.split("SERVER")[1].split("/")[0]


class DoubleWriter:
    def __init__(self, f1, f2):
        self.f1, self.f2 = f1, f2

    def write(self, msg):
        self.f1.write(msg)
        self.f2.write(msg)

    def flush(self):
        self.f1.flush()
        self.f2.flush()


class LogTracker:
    """Logging all numerical results."""

    def __init__(self):
        self._metrics = OrderedDict()
        self._print_format = {}

    def update(self, k, v):
        """ """
        if ":" in k:
            k, format_ = k.split(":")
            self._print_format[k] = format_

        if k in self._metrics:
            self._metrics[k] = np.append(self._metrics[k], v)
        else:
            self._metrics[k] = np.array([v])

    def reset(self):
        self._metrics = OrderedDict()

    def get_metrics(self, k=None):
        if k is not None:
            return self._metrics[k]
        else:
            return self._metrics

    def format(self, selected=None, not_selected=None, sep=" "):
        """Format for print."""
        logstr = []
        for k, v in self._metrics.items():
            in_selected = True

            if selected is not None:
                in_selected = False
                for s in selected.split(","):
                    if fnmatch(k, s):
                        in_selected = True
                        break

            if not_selected is not None:
                for s in not_selected.split(","):
                    if fnmatch(k, s):
                        in_selected = False
                        break

            if in_selected:
                f = self._print_format[k] if k in self._print_format else f"%s"
                logstr += [f"{k} {f}" % v[-1]]
        return sep.join(logstr)

    def get_ma(self, k, window=10):
        """Moving average."""
        return moving_average(self._metrics[k], window)


class Logger(object):
    passer = {}

    def __init__(self, args, overwrite_print=False):
        self.args = args
        self.overwrite_print = overwrite_print
        self._logtxt_name = "log.txt"

        # Load smilelogging config.
        smileconfig_file = ".smilelogging_config.yaml"
        if not os.path.exists(smileconfig_file):
            smileconfig_template = os.path.join(
                os.path.dirname(os.path.abspath(__file__)), smileconfig_file
            )
            shutil.copyfile(smileconfig_template, smileconfig_file)
        with open(smileconfig_file, "r") as f:
            smileconfig_dict = yaml.safe_load(f)
        smileconfig = SimpleNamespace(**smileconfig_dict)  # Convert dict to attributes.
        self.smileconfig = smileconfig
        self.logging_prefix = smileconfig.logging_prefix
        self._timeid_format = "%Y%m%d.%H%M%S"

        # Some default folder names.
        self._experiments_path = smileconfig_dict.get("experiments_path", "Experiments")
        self._debug_path = smileconfig_dict.get("debug_path", "Debug_Dir")

        # Handle the DDP case.
        self.global_rank, self.local_rank = self._figure_out_rank()
        self._ddp = self.global_rank >= 0

        # Get user and ip info.
        self.userip, self.hostname = self.get_userip()

        # Get various id's.
        self.pid = str(os.getpid())
        self.codeid = self._get_codeid()
        self.nodeid = "%03d" % int(self.userip.split(".")[-1])
        self.timeid = self._get_timeid()
        self.expid = self.timeid[-6:]
        self._ExpID = f"SERVER{self.nodeid}-{self.timeid}"

        # Get experiment folder name.
        self._experiment_folder = self._get_experiment_folder()

        # Create the folder structure. This can be done only after the ExpID is created.
        experiments_folder = create_folder_structure(
            experiments_path=self._debug_path if args.debug else self._experiments_path,
            experiment_folder=self._experiment_folder,
            experiment_folder_structure=smileconfig.experiment_folder_structure,
            rank=self.global_rank,
        )
        print("Creating experiment folder is done. Tree structure:")
        print_tree(experiments_folder.experiment.path)
        self.experiment_folder = experiments_folder.experiment
        self._check_folder_structure()
        self.log_path = self.experiment_folder.log.path
        self.logtxt_path = pjoin(self.experiment_folder.log.path, self._logtxt_name)
        self.sysinfo_path = self.experiment_folder.system_info.path
        self.weights_path = self.experiment_folder.weights.path
        self._cache_path = pjoin(self.experiment_folder.path, ".caches")
        self.set_up_logtxt()

        # Set up logging prefix color.
        self.logging_prefix_color_fn = self._set_up_logging_prefix_color()

        # Handle the DDP case.
        self._figure_out_rank()
        self._ddp = self.global_rank >= 0

        # Set up logging utils.
        self.log_tracker = LogTracker()

        # Initial print.
        self.print_script()
        self.print_args()

        # Cache misc environment info  (e.g., args, GPU, git, etc.) and codes.
        self.save_env_snapshot()
        if self.global_rank in [-1, 0]:  # Only the main process does caching.
            self._backup_code()

    @property
    def ExpID(self):
        warnings.warn(
            "The attribute 'ExpID' is deprecated and will be removed. Please use 'timeid' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._ExpID

    def _set_up_logging_prefix_color(self):
        """Set up log prefix color function."""
        logging_prefix_color = self.logging_prefix["color"].lower()
        if logging_prefix_color == "none":
            return None
        else:
            assert self.logging_prefix["color"] in [
                "blue",
                "green",
                "yellow",
                "red",
            ], "Invalid `color` option in `logging_prefix`. Please check!"
            return eval(logging_prefix_color)

    def _check_folder_structure(self):
        logtxt_path = pjoin(self.experiment_folder.log.path, self._logtxt_name)
        assert os.path.exists(
            logtxt_path
        ), "File `{}` not exist under {}. Please check!".format(
            self._logtxt_name, self.experiment_folder.log.path
        )
        assert os.path.exists(
            self.experiment_folder.weights.path
        ), "Folder `weights` not exist under {}. Please check!".format(
            self.experiment_folder.path
        )
        assert os.path.exists(
            self.experiment_folder.system_info.path
        ), "Folder `system_info` not exist under {}. Please check!".format(
            self.experiment_folder.path
        )
        assert os.path.exists(
            pjoin(self.experiment_folder.path, ".caches")
        ), "Folder `.caches` not exist under {}. Please check!".format(
            self.experiment_folder.path
        )

    def _get_experiment_folder(self):
        """Get the experiment folder name based on the provided template in smilelogging config file."""
        folder_name = self.smileconfig.experiment_folder["format"]
        if self._ddp:
            folder_name = self.smileconfig.experiment_folder["format_ddp"]
        if "<timeid>" in folder_name:
            folder_name = folder_name.replace("<timeid>", self.timeid)
        if "<rank>" in folder_name:
            folder_name = folder_name.replace("<rank>", str(self.global_rank))
        if "<nodeid>" in folder_name:
            folder_name = folder_name.replace("<nodeid>", self.nodeid)
        if "<experiment_name>" in folder_name:
            folder_name = folder_name.replace(
                "<experiment_name>", self.args.experiment_name
            )
        return folder_name

    def _figure_out_rank(self):
        """Figure out the process rank when using multi-processing (e.g., DDP).

        Returns:
            gloabl_rank: Process rank in all processes (across multiple nodes).
            local_rank: Process rank in the processes on the current node.
        """
        global_rank = os.getenv("GLOBAL_RANK", -1)
        local_rank = os.getenv("LOCAL_RANK", -1)
        if local_rank != -1:  # DDP is used
            if global_rank == -1:  # single-node DDP
                global_rank = local_rank
        global_rank = int(global_rank)
        local_rank = int(local_rank)

        # Get ranks from args
        if hasattr(self.args, "global_rank") and self.args.global_rank >= 0:
            global_rank = self.args.global_rank
        if hasattr(self.args, "local_rank") and self.args.local_rank >= 0:
            local_rank = self.args.local_rank

        return global_rank, local_rank

    def _get_codeid(self) -> str:
        """Return git commit ID as codeid.

        Returns:
            codeid: A string that indicates the git commit ID.
        """
        git_check = run_shell_command("git rev-parse --is-inside-work-tree")
        self.use_git = len(git_check) == 1 and git_check[0] == "true"
        self.git_root = run_shell_command("git rev-parse --show-toplevel")[
            0
        ]  # The path where .git is placed

        if self.use_git:
            codeid = run_shell_command("git rev-parse --short HEAD")[0]
            changed_files = run_shell_command("git diff --name-only")
            new_real_change = False
            for f in changed_files:
                if (
                    f and not f.endswith(".sh") and f.endswith(".slurm")
                ):  # Not consider shell script change as a formal change
                    new_real_change = True
            if new_real_change:
                logtmp = (
                    "Warning! There is (at least) one uncommitted change in your code. "
                    "Recommended to git commit it first"
                )
                self.print(logtmp)
                time.sleep(3)
        else:
            codeid = "GitNotFound"
            logtmp = "Warning! Git not found under this project. Highly recommended to use Git to manage code"
            self.print(logtmp)
        return codeid

    def _get_timeid(self) -> str:
        """Get time stamp for the experiment folder name. The expid MUST be unique.
        If multi-processing is used, only the main process can create the timeid;
        the others will simply wait for the main process to finish creating the timeid.

        Returns:
            timeid: A string that indicates the time stamp of the experiment.
        """
        time_limit = (
            120  # A process must finish creating the timeid by this time limit.
        )
        failure_msg = (
            "Hanged for over 2 mins when creating the time id, "
            "which is unusual. Please try again."
        )
        if self.global_rank in [-1, 0]:
            timeid = datetime.now(TIMEZONE).strftime(self._timeid_format)
            existing_exps = glob(f"{self._experiments_path}/*{timeid}*")
            t0 = time.time()
            # Make sure the expid is unique.
            while len(existing_exps) > 0:
                time.sleep(1)
                timeid = datetime.now(TIMEZONE).strftime(self._timeid_format)
                existing_exps = glob(f"{self._experiments_path}/*{timeid}*")
                if time.time() - t0 > time_limit:
                    self.print(failure_msg)
                    exit(1)
            return timeid
        else:
            # TODO: Corner case - if the minute is 59 and next the main process creates
            # a timeid with minute = 60, the while loop will fail.
            timeid = datetime.now(TIMEZONE).strftime(self._timeid_format)
            date_hm = timeid[:-2]
            exps = glob(f"{self._experiments_path}/*{self.args.experiment_name}*")
            exps = [e for e in exps if date_hm in e and self.nodeid in e]
            t0 = time.time()
            while len(exps) == 0:  # Main process still not created the folder.
                time.sleep(1)
                exps = glob(f"{self._experiments_path}/*{self.args.experiment_name}*")
                exps = [e for e in exps if date_hm in e and self.nodeid in e]
                if time.time() - t0 > time_limit:
                    self.print(failure_msg)
                    exit(1)
            return date_hm + exps[0].split(date_hm)[1][:2]

    def _get_experiment_path(self):
        """Get a unique directory for each experiment."""
        # Get rank for each process (used in multi-process training, e.g., DDP).
        if self.global_rank == -1:
            other_ranks_folder = ""
            rank = ""
        else:
            other_ranks_folder = "" if self.global_rank == 0 else "OtherRanks"
            rank = f"RANK{self.global_rank}-"

        experiment_path = ""

        if experiment_path != "":  # Use existing folder.
            self.ExpID = parse_ExpID(experiment_path)
        else:
            # Create a new folder.
            server = "SERVER%03d-" % int(self.userip.split(".")[-1])
            self.ExpID = rank + server + self._get_timeid()
            experiment_path = "%s/%s/%s_%s" % (
                self._experiments_path,
                other_ranks_folder,
                self.args.experiment_name,
                self.ExpID,
            )

        # --debug has the highest priority. If --debug, all the things will be saved in the debug folder.
        if self.args.debug:
            experiment_path = self._debug_path
        experiment_path = os.path.normpath(experiment_path)
        return experiment_path

    def set_up_logtxt(self):
        """Open the object of txt file and if necessary, overwrite the system print fn.

        (TODO: Overwriting the builtins.print fn may not be the best way to do so.)
        """
        self.logtxt = open(self.logtxt_path, "a+")
        if self.overwrite_print:
            sys.stderr = DoubleWriter(sys.stderr, self.logtxt)
            self.original_print = builtins.print  # Keep a copy of the original print fn
            builtins.print = self.print

    def _get_logging_prefix(self, callinfo: str):
        """Get the prefix for each line during logging."""
        prefix = self.logging_prefix["format"]
        if self.args.debug:
            prefix = self.logging_prefix["format_debug"]
        if "<time>" in prefix:
            now = datetime.now(TIMEZONE).strftime(self.logging_prefix["time_format"])
            prefix = prefix.replace("<time>", now)
        if "<pid>" in prefix:
            prefix = prefix.replace("<pid>", self.pid)
        if "<expid>" in prefix:
            prefix = prefix.replace("<expid>", self.expid)
        if "<callinfo>" in prefix:
            prefix = prefix.replace("<callinfo>", callinfo)
        if "<rank>" in prefix:
            prefix = prefix.replace("<rank>", f"{self.global_rank}")
        prefix += " $ "

        # Add color.
        if self.logging_prefix_color_fn is not None:
            prefix = self.logging_prefix_color_fn(prefix)

        return prefix

    def print(
        self,
        *msg,
        sep=" ",
        end="\n",
        file=None,
        flush=False,
        callinfo=None,
        unprefix=False,
        acc=False,
        level=0,
        main_process_only=True,
        color=None,
    ):
        """Replace the standard print func. Print to console and logtxt file."""
        if main_process_only:
            if self.global_rank not in [-1, 0]:
                return

        # Get the caller file name and line number
        if callinfo is None:
            result = traceback.extract_stack()
            caller = result[-2]
            file_path_of_caller = caller.filename
            filename = os.path.relpath(file_path_of_caller)
            if f"{os.sep}site-packages{os.sep}" in filename:
                filename = filename.split(f"{os.sep}site-packages{os.sep}")[1]
            linenum = sys._getframe().f_back.f_lineno
            callinfo = f"{filename}:{linenum}"

        # Get the level info: https://docs.python.org/3/library/logging.html#levels
        level = str(level).lower()
        assert level in [
            "0",
            "10",
            "20",
            "30",
            "40",
            "50",
            "notset",
            "debug",
            "info",
            "warning",
            "error",
            "critical",
        ]

        # Get the final message to print.
        msg = sep.join([str(m) for m in msg]) + end
        if acc:
            # Add blanks to accuracy lines for easier identification.
            msg = "  " * int(self.ExpID[-1]) + msg

        if not unprefix:
            prefix = self._get_logging_prefix(callinfo)

            # Render msg with colors.
            if color is not None:
                assert color in [
                    "green",
                    "red",
                    "blue",
                    "yellow",
                ], "Invalid `color` option. Please check!"
                color_fn = eval(color)
                msg = color_fn(msg)

            msg = prefix + msg

        # Print (TODO: Add flush?)
        if file is None:
            self.logtxt.write(msg)
            sys.stdout.write(msg)
            self.logtxt.flush()
            sys.stdout.flush()
        else:
            file.write(msg)
            file.flush()

    def info(
        self,
        *msg,
        sep=" ",
        end="\n",
        file=None,
        flush=False,
        unprefix=False,
        acc=False,
        main_process_only=True,
        color=None,
    ):
        """Reload logger.info in python logging"""
        result = traceback.extract_stack()
        caller = result[-2]  # The second last file is the caller of logger.info.
        file_path_of_caller = caller.filename
        filename = os.path.relpath(file_path_of_caller)
        if f"{os.sep}site-packages{os.sep}" in filename:
            filename = filename.split(f"{os.sep}site-packages{os.sep}")[1]
        linenum = sys._getframe().f_back.f_lineno
        callinfo = f"{filename}:{linenum}"

        self.print(
            *msg,
            sep=sep,
            end=end,
            file=file,
            flush=flush,
            callinfo=callinfo,
            unprefix=unprefix,
            acc=acc,
            level="info",
            main_process_only=main_process_only,
            color=color,
        )

    def warn(
        self,
        *msg,
        sep=" ",
        end="\n",
        file=None,
        flush=False,
        unprefix=False,
        acc=False,
        main_process_only=True,
        color=None,
    ):
        """Reload logger.warn in python logging"""
        self.print(
            *msg,
            sep=sep,
            end=end,
            file=file,
            flush=flush,
            unprefix=unprefix,
            acc=acc,
            level="warning",
            main_process_only=main_process_only,
            color="yellow",
        )

    def get_userip(self):
        user = getpass.getuser()

        # Get IP address. Refer to: https://stackoverflow.com/questions/166506/finding-local-ip-addresses-using-pythons-stdlib
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(("8.8.8.8", 80))
        ip = s.getsockname()[0]
        s.close()

        userip = f"{user}@{ip}"
        hostname = socket.gethostname()
        return userip, hostname

    def print_script(self):
        """Print the script to the start of the log txt file."""
        hostname_userip = f"hostname: {self.hostname} - userip: {self.userip} - codeid: {self.codeid} - pid: {self.pid} - timezone: {TIMEZONE}"
        cd_cmd = "cd %s" % os.path.abspath(os.getcwd())
        executable = os.path.basename(sys.executable)
        script = executable + " " + " ".join(sys.argv)
        if "CUDA_VISIBLE_DEVICES" in os.environ:
            gpu_id = os.environ["CUDA_VISIBLE_DEVICES"]
            script = f"CUDA_VISIBLE_DEVICES {gpu_id} " + script
        script = yellow("\n".join([hostname_userip, cd_cmd, script]))
        script += "\n"
        if self._ddp:
            script += yellow("Note: DDP (Data Distributed Parallel) is used.\n")
        self.print(script, unprefix=True)

    def print_args(self):
        """Example: ('batch_size', 16) ('decoder', models/small16x_ae_base/d5_base.pth).
        It will sort the arg keys in alphabeta order, case insensitive.
        """
        self.print(_pretty_dict_format(self.args.__dict__) + "\n", unprefix=True)

    def save_env_snapshot(self):
        """Save environment snapshot (such as args, gpu info, users, git info)."""
        # Save args.
        with open(pjoin(self.log_path, "args.yaml"), "w") as f:
            yaml.dump(self.args.__dict__, f, indent=2)

        # Save smilelogging config.
        with open(pjoin(self.sysinfo_path, "smilelogging_config.yaml"), "w") as f:
            yaml.dump(self.smileconfig.__dict__, f, indent=2)

        # Save system info.
        if check_command_installed("nvidia-smi"):
            os.system(
                "nvidia-smi >> {}".format(pjoin(self.sysinfo_path, "nvidia-smi.txt"))
            )
        if check_command_installed("gpustat"):
            os.system("gpustat >> {}".format(pjoin(self.sysinfo_path, "gpustat.txt")))
        if check_command_installed("who"):
            os.system("who -b >> {}".format(pjoin(self.sysinfo_path, "who.txt")))
            os.system("who >> {}".format(pjoin(self.sysinfo_path, "who.txt")))

        # Save git info.
        if self.use_git:
            os.system(
                "git status >> {}".format(pjoin(self.sysinfo_path, "git_status.txt"))
            )

    def plot(self, name, out_path):
        self.log_tracker.plot(name, out_path)

    def _backup_code(self):
        """Back up the code. Do not back up in debug mode, or, explicitly indicated."""
        cache_code = self.smileconfig.cache_code
        if self.args.debug or not cache_code["is_open"]:
            return

        cache_script = os.path.expanduser(cache_code["script"])  # In case ~ is used.
        if os.path.exists(cache_script):
            t0 = time.time()
            logtmp = f"==> Caching code to {bold(self._cache_path)} using the provided script {bold(cache_script)}"
            self.print(logtmp)
            cmd = f"sh {cache_script} {self._cache_path} > /dev/null"
            os.system(cmd)
            logtmp = f"==> Caching done (time: {time.time() - t0:.2f}s)"
            self.print(logtmp)

    def print_model_arch(self, model):
        """Print model architecture to log."""
        with open(pjoin(self.log_path, "model_arch.txt"), "w") as f:
            print("%s\n" % str(model), file=f, flush=True)

    def timenow(self):
        return datetime.now(TIMEZONE).strftime("%Y/%m/%d-%H:%M:%S")
